#define _GNU_SOURCE
#define _SVID_SOURCE
#include <nss.h>
#include <stdio.h>
#include <pwd.h>
#include <grp.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdbool.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <dirent.h>

/* code generators */
#include "cgen.h"

#include "buf.h"
#include "nat.h"
#include "list.h"

/* constants */
#include "config/shell.h"
#include "config/debug.h"
#include "config/spool.h"

static unsigned char g_uids[65536];
static unsigned short g_uid = 0;

/* I'm going to assume that every appropriately named dirent represents a user.
 * We may need to revise this, e.g. by checking for the existence of home dir.
 * We're also going to be evil and use the scandir() traversal to directly
 * record allocated uids. <MS> */
int
uid_filter(const struct dirent* d) {
  unsigned long val;
  if(parse_nat(&val, d->d_name, strlen(d->d_name)) == 1) return 0;
  if(val < 10000 || val > 60000) return 0;
  g_uids[val] = 1;
  return 0;
}

/* implementation */
enum nss_status
_nss_rainbow_endpwent(void) {
  return NSS_STATUS_SUCCESS;
}

enum nss_status
_nss_rainbow_setpwent(void) {
  for (unsigned short i = 10000; i < 60000; i++)
    g_uids[i] = 0;
  g_uid = 0;
  struct dirent** uids;
  CHK(scandir(SPOOL "/uid_pool", &uids, uid_filter, versionsort) == -1,
      "Unable to extract uids from $RAINBOW_SPOOL.", out_error_spool);
  return NSS_STATUS_SUCCESS;
out_error_spool:
  return NSS_STATUS_UNAVAIL;
}

enum nss_status
_nss_rainbow_getpwent_r(struct passwd *result, char* buf, size_t buflen, int *errnop) {
  /* I'm going to interpret the reentrancy requirement on this function to mean
   * that the sequence of all results returned by all calls to this function
   * between calls to setpwent() should contain contain every entry exactly
   * once. It is undefined whether writes to the underlying storage that
   * interleave with calls to getpwent() will be visible to callers. */

  char gid_path[NAME_MAX+1];
  char gid_name[NAME_MAX+1];
  size_t gid_path_len;
  ssize_t gid_name_len;
  char* gid_path_str;
  unsigned long gid;

begin:
  gid_path_len = gid_name_len = NAME_MAX+1;
  gid_path_str = gid_path;

  /* Locate the next reserved uid. */
  while (true){
    g_uid++;
    if (!g_uid) return NSS_STATUS_UNAVAIL;
    if (g_uids[g_uid]) break;
  }

  result->pw_uid = g_uid;

  CHK(format_buf(&gid_path_str, &gid_path_len, SPOOL "/uid_to_gid/%d", g_uid) == 1,
      "Unable to calculate gid-path.", begin);
  LET(gid_name_len = readlink(gid_path, gid_name, gid_name_len), gid_name_len == -1,
      "Unable to read gid-path.", begin);
  CHK(parse_nat(&gid, gid_name, gid_name_len) == 1,
      "Unable to parse gid_buf into a gid.", begin);
  CHK(gid < 10000 || gid > 60000,
      "gid not in valid range.", begin);
  result->pw_gid = gid;

  result->pw_dir = buf;
  CHK(format_buf(&buf, &buflen, SPOOL "/uid_to_home_dir/%d", g_uid) == 1,
      "Unable to calculate home dir.", begin);

  result->pw_gecos = result->pw_name = result->pw_passwd = buf;
  CHK(format_buf(&buf, &buflen, "%d", g_uid) == 1,
      "Not enough buffer for $USER.", begin);

  result->pw_shell = buf;
  CHK(write_buf(&buf, &buflen, SHELL) == 1,
      "Shell string constant too long.", begin);

  if(getenv(DEBUG) != NULL)
    dprintf(2, "%s success: %s (%d, %d) %s %s %s\n", __func__,
            result->pw_name, result->pw_uid, result->pw_gid, result->pw_dir,
            result->pw_shell, result->pw_passwd);

  return NSS_STATUS_SUCCESS;
}

enum nss_status _nss_rainbow_getpwuid_r(uid_t uid, struct passwd *result, char* buf, size_t buflen, int *errnop) {
  if (uid < 10000)
      return NSS_STATUS_NOTFOUND;

  result->pw_dir = buf;
  CHK(format_buf(&buf, &buflen, SPOOL "/uid_to_home_dir/%d", uid) == 1,
      "Unable to calculate home dir.", out_error_errno);

  struct stat st;
  CHK(stat(result->pw_dir, &st) == -1,
      "Stat failed for homebuf.", out_error_errno);
  result->pw_uid = uid;
  result->pw_gid = st.st_gid;

  result->pw_gecos = result->pw_name = result->pw_passwd = buf;
  CHK(format_buf(&buf, &buflen, "%d", uid) == 1,
      "Unable to calculate user name.", out_error_errno);

  result->pw_shell = buf;
  CHK(write_buf(&buf, &buflen, SHELL) == 1,
      "Shell string constant too long.", out_error_errno);

  if(getenv(DEBUG) != NULL)
    dprintf(2, "%s success in uid: %s (%d, %d) %s %s %s\n", __func__,
            result->pw_name, result->pw_uid, result->pw_gid,
            result->pw_dir, result->pw_shell, result->pw_passwd);

  return NSS_STATUS_SUCCESS;

out_error_errno:
  *errnop = errno;
  return NSS_STATUS_TRYAGAIN;
}

enum nss_status _nss_rainbow_getpwnam_r(const char * name, struct passwd *result, char* buf, size_t buflen, int *errnop) {
  unsigned long val;
  CHK(parse_nat(&val, name, strlen(name)) == 1,
      "Unable to parse name into a uid.", out_error_name);

  TST(val < 10000 || val > 60000, EINVAL,
      "Name specifies a uid outside [10000, 60000].", out_error_name);

  return _nss_rainbow_getpwuid_r((uid_t)val, result, buf, buflen, errnop);

out_error_name:
  *errnop = errno;
  return NSS_STATUS_NOTFOUND;
}

#if 0
enum nss_status
_nss_rainbow_initgroups_dyn(const char *user, gid_t gid, long int *start, long int *size, gid_t **groupsp, long int limit, int *errnop) {
  if (uid < 10000)
      return NSS_STATUS_NOTFOUND;

  int pathlen = 1024;
  char pathdata[1024], *pathbuf = pathdata;

  result->pw_dir = buf;
  CHK(format_buf(&pathbuf, &pathlen, SPOOL "/uid_to_home_dir/%d", uid) == 1,
      "Unable to calculate home dir.", out_error_errno);

  struct stat st;
  CHK(stat(result->pw_dir, &st) == -1,
      "Stat failed for homebuf.", out_error_errno);
  result->pw_uid = uid;
  result->pw_gid = st.st_gid;

  (*groupsp)[*start] =

  return NSS_STATUS_SUCCESS;

out_error_errno:
  *ernnop = errno;
  return NSS_STATUS_NOTFOUND;
}

#endif

struct uid_list {
  struct list_head list;
  uid_t uid;
};

struct gid_list {
  struct list_head list;
  struct uid_list uids;
  gid_t gid;
};

struct gid_list g_gids;
struct gid_list* g_gid;

/* I'm going to assume that every appropriately named dirent represents a user.
 * We may need to revise this, e.g. by checking for the existence of home dir.
 * We're also going to be evil and use the scandir() traversal to directly
 * record allocated uids. <MS> */
int
uid_to_gid_filter(const struct dirent* d) {
  unsigned long uid, gid;
  char gid_path[NAME_MAX+1];
  char gid_name[NAME_MAX+1];
  size_t gid_path_len = NAME_MAX + 1;
  ssize_t gid_name_len = NAME_MAX + 1;
  char* gid_path_str = gid_path;

  CHK(parse_nat(&uid, d->d_name, strlen(d->d_name)) == 1,
      "Unable to parse uid.", exit);
  CHK(uid < 10000 || uid > 60000,
      "uid not in valid range.", exit);

  CHK(format_buf(&gid_path_str, &gid_path_len, SPOOL "/uid_to_gid/%d", uid) == 1,
      "Unable to calculate uid-to-gid-path.", exit);
  LET(gid_name_len = readlink(gid_path, gid_name, gid_name_len), gid_name_len == -1,
      "Unable to read gid-path.", exit);
  CHK(parse_nat(&gid, gid_name, gid_name_len) == 1,
      "Unable to parse gid_buf into a gid.", exit);
  CHK(gid < 10000 || gid > 60000,
      "gid not in valid range.", exit);

  /* We might have already allocated a uid_list for our gid.
   * Therefore, search for it. */
  struct gid_list* gid_iter = NULL;
  list_for_each_entry(gid_iter, &g_gids.list, list) {
    if (gid_iter->gid == gid)
      break;
  }

  /* If we find no uid_list, make one and insert it. */
  if (&gid_iter->list == &g_gids.list) {
    LET(gid_iter = malloc(sizeof(struct gid_list)), !gid_iter,
        "Unable to allocate new gid_list node.", exit);
    gid_iter->gid = gid;
    INIT_LIST_HEAD(&(gid_iter->uids.list));
    list_add_tail(&(gid_iter->list), &g_gids.list);
    g_gid = gid_iter;
  }

  /* gid_iter now points to the appropriate gid_list node.
   * There's no way that we could already have our current (uid, gid) row.
   * Therefore, we'll immediately insert it. */
  struct uid_list* uid_iter;
  LET(uid_iter = malloc(sizeof(struct uid_list)), !uid_iter,
      "Unable to allocate new uid_list node.", exit);
  uid_iter->uid = uid;
  list_add_tail(&(uid_iter->list), &(gid_iter->uids.list));
exit:
  return 0;
}


enum nss_status _nss_rainbow_setgrent(void) {
  struct dirent** uids;
  INIT_LIST_HEAD(&g_gids.list);
  INIT_LIST_HEAD(&g_gids.uids.list);
  g_gid = &g_gids;
  CHK(scandir(SPOOL "/uid_to_gid", &uids, uid_to_gid_filter, versionsort) == -1,
      "Unable to extract uid->gids links from $RAINBOW_SPOOL.", out_error_spool);
  g_gid = list_entry(g_gids.list.next, struct gid_list, list);
  return NSS_STATUS_SUCCESS;
out_error_spool:
  return NSS_STATUS_UNAVAIL;
}

enum nss_status _nss_rainbow_endgrent(void) {
  struct gid_list* gid_iter;
  struct uid_list* uid_iter;
  while(!list_empty(&g_gids.list)) {
    gid_iter = list_entry(g_gids.list.next, struct gid_list, list);
    while(!list_empty(&(gid_iter->uids.list))) {
      uid_iter = list_entry(gid_iter->uids.list.next, struct uid_list, list);
      list_del(&(uid_iter->list));
      free(uid_iter);
    }
    list_del(&(gid_iter->list));
    free(gid_iter);
  }
  return NSS_STATUS_SUCCESS;
}
/*
struct group
{
  char *gr_name;
  char *gr_passwd;
  __gid_t gr_gid;
  char **gr_mem;
};
*/

int
fill_group(struct group* result, char* buf, size_t buflen, int *errnop, struct gid_list* gid_iter) {
  result->gr_gid = gid_iter->gid;

  result->gr_name = result->gr_passwd = buf;
  TST(format_buf(&buf, &buflen, "%d", gid_iter->gid) == 1, ENOMEM,
      "Unable to calculate group name.", out_error_errno);

  /* count the number of members */
  struct uid_list* uid_iter;
  struct list_head* uid_node_iter;
  size_t member_count = 0;
  list_for_each(uid_node_iter, &(gid_iter->uids.list)) {
    member_count++;
  }

  /* reserve space for the member-pointers */
  size_t member_size = (member_count+1)* sizeof(char*);
  result->gr_mem = buf;
  TST(buflen < member_size, EINVAL,
      "Not enough space for group member pointers.", out_error_errno);
  result->gr_mem[member_count] = 0;
  buf += member_size;
  buflen -= member_size;

  /* fill in the members and swing the pointers */
  member_count = 0;
  list_for_each_entry(uid_iter, &(gid_iter->uids.list), list) {
    result->gr_mem[member_count] = buf;
    CHK(format_buf(&buf, &buflen, "%d", uid_iter->uid) == 1,
        "Unable to write a group member.", out_error_errno);
    member_count++;
  }

  return 0;

out_error_errno:
  *errnop = errno;
  return 1;
}

enum nss_status
_nss_rainbow_getgrent_r(struct group *result, char *buf, size_t buflen, int *errnop) {
  if (&g_gid->list == &g_gids.list)
    return NSS_STATUS_UNAVAIL;

  CHK(fill_group(result, buf, buflen, errnop, g_gid) == 1,
      "Unable to fill in group struct.", out_error_errno);
  g_gid = list_entry(g_gid->list.next, struct gid_list, list);

  return NSS_STATUS_SUCCESS;

out_error_errno:
  *errnop = errno;
  return NSS_STATUS_TRYAGAIN;
}


enum nss_status
_nss_rainbow_getgrgid_r(gid_t gid, struct group *result, char *buf, size_t buflen, int *errnop) {
  enum nss_status ret;
  if (gid < 10000 || gid > 60000)
      return NSS_STATUS_NOTFOUND;

  /* XXX CHECK RETURN CODE */
  _nss_rainbow_setgrent();

  /* look up group data */
  struct gid_list* gid_iter = NULL;
  list_for_each_entry(gid_iter, &g_gids.list, list) {
    if (gid_iter->gid == gid)
      break;
  }
  if (&gid_iter->list == &g_gids.list)
  {
    ret = NSS_STATUS_NOTFOUND;
    goto out_dealloc;
  }

  if(fill_group(result, buf, buflen, errnop, gid_iter) == 1)
  {
      PERROR("Unable to fill in group struct.");
      goto out_error;
  }

  if(getenv(DEBUG) != NULL)
    dprintf(2, "%s success in gid: %s (%d) %s\n", __func__,
            result->gr_name, result->gr_gid, result->gr_passwd);

  ret = NSS_STATUS_SUCCESS;
  goto out_dealloc;

out_error:
  *errnop = errno;
out_dealloc:
  /* XXX CHECK RETURN CODE */
  _nss_rainbow_endgrent();
  return ret;
}

enum nss_status
_nss_rainbow_getgrnam_r(const char* name, struct group *result, char *buf, size_t buflen, int *errnop) {
  unsigned long val;
  CHK(parse_nat(&val, name, strlen(name)) == 1,
      "Unable to parse name into a gid.", out_error_name);
  TST(val < 10000 || val > 60000, EINVAL,
      "Name specifies a gid outside [10000, 60000].", out_error_name);
  return _nss_rainbow_getgrgid_r((gid_t)val, result, buf, buflen, errnop);
out_error_name:
  *errnop = errno;
  return NSS_STATUS_NOTFOUND;
}
